@@ -16,15 +16,15 @@
 from email import header
 from email.mime import multipart
 from email.mime import text
-import json
 import smtplib
 import time
+from urllib import parse
 
 from oslo_log import log as logging
 import requests
-import six
 
 from mistral import exceptions as exc
+from mistral import utils
 from mistral.utils import javascript
 from mistral.utils import ssh_utils
 from mistral_lib import actions
@@ -42,19 +42,40 @@ class EchoAction(actions.Action):
     stub.
     """
 
-    def __init__(self, output):
+    def __init__(self, output, delay=0):
         super(EchoAction, self).__init__()
 
         self.output = output
 
+        try:
+            self._delay = float(delay)
+            self._delay = 0 if self._delay < 0 else self._delay
+        except ValueError:
+            self._delay = 0
+
     def run(self, context):
-        LOG.info('Running echo action [output=%s]', self.output)
+        LOG.info(
+            'Running echo action [output=%s, delay=%s]',
+            self.output,
+            self._delay
+        )
+
+        time.sleep(self._delay)
 
         return self.output
 
     def test(self, context):
         return 'Echo'
 
+    def __eq__(self, other):
+        if type(other) is not EchoAction:
+            return False
+
+        if self.output != other.output or self._delay != other._delay:
+            return False
+
+        return True
+
 
 class NoOpAction(actions.Action):
     """No-operation action.
@@ -95,6 +116,9 @@ class FailAction(actions.Action):
 
         self.error_data = error_data
 
+    def __repr__(self):
+        return 'Fail action'
+
     def run(self, context):
         LOG.info('Running fail action.')
 
@@ -120,6 +144,8 @@ class HTTPAction(actions.Action):
         query string for the HTTP request.
     :param body: (optional) Dictionary, bytes, or file-like object to send
         in the body of the HTTP request.
+    :param json: (optional) A JSON serializable Python object to send
+        in the body of the HTTP request.
     :param headers: (optional) Dictionary of HTTP Headers to send with
         the HTTP request.
     :param cookies: (optional) Dict or CookieJar object to send with
@@ -141,6 +167,7 @@ class HTTPAction(actions.Action):
                  method="GET",
                  params=None,
                  body=None,
+                 json=None,
                  headers=None,
                  cookies=None,
                  auth=None,
@@ -157,13 +184,19 @@ class HTTPAction(actions.Action):
 
         if isinstance(headers, dict):
             for key, val in headers.items():
-                if isinstance(val, (six.integer_types, float)):
+                if isinstance(val, (int, float)):
                     headers[key] = str(val)
 
+        if body and json:
+            raise exc.ActionException(
+                "Only one of the parameters 'json' and 'body' can be passed"
+            )
+
         self.url = url
         self.method = method
         self.params = params
-        self.body = json.dumps(body) if isinstance(body, dict) else body
+        self.body = utils.to_json_str(body) if isinstance(body, dict) else body
+        self.json = json
         self.headers = headers
         self.cookies = cookies
         self.timeout = timeout
@@ -174,13 +207,14 @@ class HTTPAction(actions.Action):
     def run(self, context):
         LOG.info(
             "Running HTTP action "
-            "[url=%s, method=%s, params=%s, body=%s, headers=%s,"
-            " cookies=%s, auth=%s, timeout=%s, allow_redirects=%s,"
-            " proxies=%s, verify=%s]",
+            "[url=%s, method=%s, params=%s, body=%s, json=%s,"
+            " headers=%s, cookies=%s, auth=%s, timeout=%s,"
+            " allow_redirects=%s, proxies=%s, verify=%s]",
             self.url,
             self.method,
             self.params,
             self.body,
+            self.json,
             self.headers,
             self.cookies,
             self.auth,
@@ -191,7 +225,7 @@ class HTTPAction(actions.Action):
         )
 
         try:
-            url_data = six.moves.urllib.parse.urlsplit(self.url)
+            url_data = parse.urlsplit(self.url)
             if 'https' == url_data.scheme:
                 action_verify = self.verify
             else:
@@ -202,6 +236,7 @@ class HTTPAction(actions.Action):
                 self.url,
                 params=self.params,
                 data=self.body,
+                json=self.json,
                 headers=self.headers,
                 cookies=self.cookies,
                 auth=self.auth,
@@ -225,8 +260,8 @@ class HTTPAction(actions.Action):
 
         # Represent important resp data as a dictionary.
         try:
-            content = resp.json(encoding=resp.encoding)
-        except Exception as e:
+            content = resp.json()
+        except Exception:
             LOG.debug("HTTP action response is not json.")
             content = resp.content
             if content and resp.encoding not in (None, 'utf-8'):
@@ -278,9 +313,9 @@ class MistralHTTPAction(HTTPAction):
 
 
 class SendEmailAction(actions.Action):
-    def __init__(self, from_addr, to_addrs, smtp_server, cc_addrs=None,
-                 bcc_addrs=None, smtp_password=None, subject=None, body=None,
-                 html_body=None):
+    def __init__(self, from_addr, to_addrs, smtp_server, reply_to=None,
+                 cc_addrs=None, bcc_addrs=None, smtp_password=None,
+                 subject=None, body=None, html_body=None):
         super(SendEmailAction, self).__init__()
         # TODO(dzimine): validate parameters
 
@@ -288,6 +323,7 @@ class SendEmailAction(actions.Action):
         self.to = to_addrs
         self.cc = cc_addrs or []
         self.bcc = bcc_addrs or []
+        self.reply_to = reply_to or []
         self.subject = subject or "<No subject>"
         self.body = body or "<No body>"
         self.html_body = html_body
@@ -300,10 +336,11 @@ class SendEmailAction(actions.Action):
     def run(self, context):
         LOG.info(
             "Sending email message "
-            "[from=%s, to=%s, cc=%s, bcc=%s, subject=%s, using smtp=%s, "
-            "body=%s...]",
+            "[from=%s, to=%s, reply_to=%s, cc=%s, bcc=%s, subject=%s, "
+            "using smtp=%s, body=%s...]",
             self.sender,
             self.to,
+            self.reply_to,
             self.cc,
             self.bcc,
             self.subject,
@@ -322,6 +359,7 @@ class SendEmailAction(actions.Action):
                                          _charset='utf-8'))
         message['Subject'] = header.Header(self.subject, 'utf-8')
         message['From'] = self.sender
+        message['Reply-To'] = header.Header(', '.join(self.reply_to))
         message['To'] = ', '.join(self.to)
 
         if self.cc:
@@ -351,10 +389,11 @@ class SendEmailAction(actions.Action):
         # to return a result.
         LOG.info(
             "Sending email message "
-            "[from=%s, to=%s, cc=%s, bcc=%s, subject=%s, using smtp=%s, "
-            "body=%s...]",
+            "[from=%s, to=%s, reply_to=%s, cc=%s, bcc=%s, subject=%s, "
+            " using smtp=%s, body=%s...]",
             self.sender,
             self.to,
+            self.reply_to,
             self.cc,
             self.bcc,
             self.subject,
@@ -376,7 +415,7 @@ class SSHAction(actions.Action):
         return ssh_utils.execute_command
 
     def __init__(self, cmd, host, username,
-                 password="", private_key_filename=None):
+                 password="", private_key_filename=None, private_key=None):
         super(SSHAction, self).__init__()
 
         self.cmd = cmd
@@ -384,21 +423,26 @@ class SSHAction(actions.Action):
         self.username = username
         self.password = password
         self.private_key_filename = private_key_filename
+        self.private_key = private_key
 
         self.params = {
             'cmd': self.cmd,
             'host': self.host,
             'username': self.username,
             'password': self.password,
-            'private_key_filename': self.private_key_filename
+            'private_key_filename': self.private_key_filename,
+            'private_key': self.private_key
         }
 
     def run(self, context):
         def raise_exc(parent_exc=None):
             message = ("Failed to execute ssh cmd "
                        "'%s' on %s" % (self.cmd, self.host))
+            # We suppress the actual parent error messages in favor of
+            # more generic ones as we might be leaking information to the CLI
             if parent_exc:
-                message += "\nException: %s" % str(parent_exc)
+                # The full error message needs to be logged regardless
+                LOG.exception(message + " Exception: %s", str(parent_exc))
             raise exc.ActionException(message)
 
         try:
@@ -425,7 +469,7 @@ class SSHAction(actions.Action):
             return raise_exc(parent_exc=e)
 
     def test(self, context):
-        return json.dumps(self.params)
+        return utils.to_json_str(self.params)
 
 
 class SSHProxiedAction(SSHAction):
@@ -435,13 +479,15 @@ class SSHProxiedAction(SSHAction):
 
     def __init__(self, cmd, host, username, private_key_filename,
                  gateway_host, gateway_username=None,
-                 password=None, proxy_command=None):
+                 password=None, proxy_command=None,
+                 private_key=None):
         super(SSHProxiedAction, self).__init__(
             cmd,
             host,
             username,
             password,
-            private_key_filename
+            private_key_filename,
+            private_key
         )
 
         self.gateway_host = gateway_host
