@@ -32,15 +32,21 @@ LOG = logging.getLogger(__name__)
 class CronTriggersController(rest.RestController):
     @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.CronTrigger, wtypes.text)
-    def get(self, name):
-        """Returns the named cron_trigger."""
+    def get(self, identifier):
+        """Returns the named cron_trigger.
+
+        :param identifier: Id or name of cron trigger to retrieve
+        """
         acl.enforce('cron_triggers:get', context.ctx())
 
-        LOG.info('Fetch cron trigger [name=%s]' % name)
+        LOG.debug('Fetch cron trigger [identifier=%s]', identifier)
 
-        db_model = db_api.get_cron_trigger(name)
+        # Use retries to prevent possible failures.
+        db_model = rest_utils.rest_retry_on_db_error(
+            db_api.get_cron_trigger
+        )(identifier)
 
-        return resources.CronTrigger.from_dict(db_model.to_dict())
+        return resources.CronTrigger.from_db_model(db_model)
 
     @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(
@@ -52,49 +58,58 @@ class CronTriggersController(rest.RestController):
         """Creates a new cron trigger.
 
         :param cron_trigger: Required. Cron trigger structure.
-
         """
         acl.enforce('cron_triggers:create', context.ctx())
 
-        LOG.info('Create cron trigger: %s' % cron_trigger)
+        LOG.debug('Create cron trigger: %s', cron_trigger)
 
         values = cron_trigger.to_dict()
 
-        db_model = triggers.create_cron_trigger(
-            values['name'],
-            values.get('workflow_name'),
-            values.get('workflow_input'),
-            values.get('workflow_params'),
-            values.get('pattern'),
-            values.get('first_execution_time'),
-            values.get('remaining_executions'),
+        db_model = rest_utils.rest_retry_on_db_error(
+            triggers.create_cron_trigger
+        )(
+            name=values['name'],
+            workflow_name=values.get('workflow_name'),
+            workflow_input=values.get('workflow_input'),
+            workflow_params=values.get('workflow_params'),
+            pattern=values.get('pattern'),
+            first_time=values.get('first_execution_time'),
+            count=values.get('remaining_executions'),
             workflow_id=values.get('workflow_id')
         )
 
-        return resources.CronTrigger.from_dict(db_model.to_dict())
+        return resources.CronTrigger.from_db_model(db_model)
 
     @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(None, wtypes.text, status_code=204)
-    def delete(self, name):
-        """Delete cron trigger."""
+    def delete(self, identifier):
+        """Delete cron trigger.
+
+        :param identifier: Id or name of cron trigger to delete
+        """
         acl.enforce('cron_triggers:delete', context.ctx())
 
-        LOG.info("Delete cron trigger [name=%s]" % name)
+        LOG.debug("Delete cron trigger [identifier=%s]", identifier)
 
-        db_api.delete_cron_trigger(name)
+        rest_utils.rest_retry_on_db_error(
+            triggers.delete_cron_trigger
+        )(identifier)
 
+    @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.CronTriggers, types.uuid, int,
                          types.uniquelist, types.list, types.uniquelist,
                          wtypes.text, wtypes.text, types.uuid, types.jsontype,
                          types.jsontype, resources.SCOPE_TYPES, wtypes.text,
                          wtypes.IntegerType(minimum=1), wtypes.text,
-                         wtypes.text, wtypes.text, wtypes.text)
+                         wtypes.text, wtypes.text, wtypes.text,
+                         types.uuid, bool)
     def get_all(self, marker=None, limit=None, sort_keys='created_at',
                 sort_dirs='asc', fields='', name=None, workflow_name=None,
                 workflow_id=None, workflow_input=None, workflow_params=None,
                 scope=None, pattern=None, remaining_executions=None,
                 first_execution_time=None, next_execution_time=None,
-                created_at=None, updated_at=None):
+                created_at=None, updated_at=None, project_id=None,
+                all_projects=False):
         """Return all cron triggers.
 
         :param marker: Optional. Pagination marker for large data sets.
@@ -124,6 +139,8 @@ class CronTriggersController(rest.RestController):
         :param pattern: Optional. Keep only resources with a specific pattern.
         :param remaining_executions: Optional. Keep only resources with a
                                      specific number of remaining executions.
+        :param project_id: Optional. Keep only resources with the specific
+                           project id.
         :param first_execution_time: Optional. Keep only resources with a
                                      specific time and date of first execution.
         :param next_execution_time: Optional. Keep only resources with a
@@ -132,9 +149,13 @@ class CronTriggersController(rest.RestController):
                            time and date.
         :param updated_at: Optional. Keep only resources with specific latest
                            update time and date.
+        :param all_projects: Optional. Get resources of all projects.
         """
         acl.enforce('cron_triggers:list', context.ctx())
 
+        if all_projects:
+            acl.enforce('cron_triggers:list:all_projects', context.ctx())
+
         filters = filter_utils.create_filters_from_request_params(
             created_at=created_at,
             name=name,
@@ -147,13 +168,14 @@ class CronTriggersController(rest.RestController):
             pattern=pattern,
             remaining_executions=remaining_executions,
             first_execution_time=first_execution_time,
-            next_execution_time=next_execution_time
+            next_execution_time=next_execution_time,
+            project_id=project_id,
         )
 
-        LOG.info(
+        LOG.debug(
             "Fetch cron triggers. marker=%s, limit=%s, sort_keys=%s, "
-            "sort_dirs=%s, filters=%s",
-            marker, limit, sort_keys, sort_dirs, filters
+            "sort_dirs=%s, filters=%s, all_projects=%s",
+            marker, limit, sort_keys, sort_dirs, filters, all_projects
         )
 
         return rest_utils.get_all(
@@ -166,5 +188,6 @@ class CronTriggersController(rest.RestController):
             sort_keys=sort_keys,
             sort_dirs=sort_dirs,
             fields=fields,
+            all_projects=all_projects,
             **filters
         )
