@@ -25,6 +25,12 @@ eventlet.monkey_patch(
     socket=True,
     thread=False if '--use-debugger' in sys.argv else True,
     time=True)
+# Monkey patch the original current_thread to use the up-to-date _active
+# global variable. See https://bugs.launchpad.net/bugs/1863021 and
+# https://github.com/eventlet/eventlet/issues/592
+import __original_module_threading as orig_threading
+import threading  # noqa
+orig_threading.current_thread.__globals__['_active'] = threading._active
 
 
 import os
@@ -32,9 +38,14 @@ import os
 
 # If ../mistral/__init__.py exists, add ../ to Python search path, so that
 # it will override what happens to be installed in /usr/(local/)lib/python...
-POSSIBLE_TOPDIR = os.path.normpath(os.path.join(os.path.abspath(sys.argv[0]),
-                                   os.pardir,
-                                   os.pardir))
+POSSIBLE_TOPDIR = os.path.normpath(
+    os.path.join(
+        os.path.abspath(sys.argv[0]),
+        os.pardir,
+        os.pardir
+    )
+)
+
 if os.path.exists(os.path.join(POSSIBLE_TOPDIR, 'mistral', '__init__.py')):
     sys.path.insert(0, POSSIBLE_TOPDIR)
 
@@ -65,7 +76,9 @@ def launch_thread(server, workers=1):
 
         if not SERVER_THREAD_MANAGER:
             SERVER_THREAD_MANAGER = service.ServiceLauncher(
-                CONF, restart_method='mutate')
+                CONF,
+                restart_method='mutate'
+            )
 
         SERVER_THREAD_MANAGER.launch_service(server, workers=workers)
     except Exception as e:
@@ -79,7 +92,9 @@ def launch_process(server, workers=1):
 
         if not SERVER_PROCESS_MANAGER:
             SERVER_PROCESS_MANAGER = service.ProcessLauncher(
-                CONF, restart_method='mutate')
+                CONF,
+                restart_method='mutate'
+            )
 
         SERVER_PROCESS_MANAGER.launch_service(server, workers=workers)
     except Exception as e:
@@ -116,12 +131,12 @@ def launch_any(options):
     global SERVER_PROCESS_MANAGER
     global SERVER_THREAD_MANAGER
 
-    if SERVER_PROCESS_MANAGER:
-        SERVER_PROCESS_MANAGER.wait()
-
     if SERVER_THREAD_MANAGER:
         SERVER_THREAD_MANAGER.wait()
 
+    if SERVER_PROCESS_MANAGER:
+        SERVER_PROCESS_MANAGER.wait()
+
 
 # Map cli options to appropriate functions. The cli options are
 # registered in mistral's config.py.
@@ -173,7 +188,9 @@ def get_properly_ordered_parameters():
             else:
                 conf_file_value = args[args.index(arg) + 1]
                 args.remove(conf_file_value)
+
             args.remove(arg)
+
             args.insert(0, "--config-file")
             args.insert(1, conf_file_value)
 
@@ -184,9 +201,11 @@ def override_keystone_options():
     # TODO(wxy): This function is used for keeping backward compatibility.
     # Remove it in Stein.
     auth_opts = CONF['keystone_authtoken']
+
     for opt, value in auth_opts.items():
         if opt in CONF['keystone']:
             default_value = auth_opts._group._opts[opt]['opt'].default
+
             if default_value != value != CONF['keystone'][opt]:
                 LOG.warning("The config option '%s' in section "
                             "[keystone_authtoken] has the same copy in "
@@ -197,10 +216,14 @@ def override_keystone_options():
 
 def main():
     try:
+        CONF.register_cli_opts(config.CLI_OPTS)
+
         config.parse_args(get_properly_ordered_parameters())
+
         print_server_info()
 
         logging.setup(CONF, 'Mistral')
+
         override_keystone_options()
 
         # Please refer to the oslo.messaging documentation for transport
@@ -226,8 +249,10 @@ def main():
         else:
             # Validate launch option.
             if set(cfg.CONF.server) - set(LAUNCH_OPTIONS.keys()):
-                raise Exception('Valid options are all or any combination of '
-                                ', '.join(LAUNCH_OPTIONS.keys()))
+                raise Exception(
+                    "Valid options are 'all' or any combination of [%s]" %
+                    ', '.join(LAUNCH_OPTIONS.keys())
+                )
 
             # Launch distinct set of server(s).
             launch_any(set(cfg.CONF.server))
@@ -241,6 +266,7 @@ def main():
 def reset_server_managers():
     global SERVER_THREAD_MANAGER
     global SERVER_PROCESS_MANAGER
+
     SERVER_THREAD_MANAGER = None
     SERVER_PROCESS_MANAGER = None
 
@@ -248,12 +274,14 @@ def reset_server_managers():
 # Helper method used in unit tests to access the service launcher.
 def get_server_thread_manager():
     global SERVER_THREAD_MANAGER
+
     return SERVER_THREAD_MANAGER
 
 
 # Helper method used in unit tests to access the process launcher.
 def get_server_process_manager():
     global SERVER_PROCESS_MANAGER
+
     return SERVER_PROCESS_MANAGER
 
 
