@@ -19,7 +19,6 @@ import contextlib
 from oslo_config import cfg
 from oslo_db import api as db_api
 
-
 _BACKEND_MAPPING = {
     'sqlalchemy': 'mistral.db.v2.sqlalchemy.api',
 }
@@ -171,6 +170,82 @@ def delete_workflow_definitions(**kwargs):
     IMPL.delete_workflow_definitions(**kwargs)
 
 
+# Dynamic actions.
+
+def get_dynamic_action_definition(identifier, namespace='', fields=()):
+    return IMPL.get_dynamic_action_definition(identifier, fields, namespace)
+
+
+def load_dynamic_action_definition(identifier, namespace='', fields=()):
+    return IMPL.load_dynamic_action_definition(identifier, fields, namespace)
+
+
+def create_dynamic_action_definition(values):
+    return IMPL.create_dynamic_action_definition(values)
+
+
+def update_dynamic_action_definition(identifier, values, namespace=''):
+    return IMPL.update_dynamic_action_definition(identifier, values, namespace)
+
+
+def get_dynamic_action_definitions(limit=None, marker=None, sort_keys=None,
+                                   sort_dirs=None, fields=None, **kwargs):
+    return IMPL.get_dynamic_action_definitions(
+        limit=limit,
+        marker=marker,
+        sort_keys=sort_keys,
+        sort_dirs=sort_dirs,
+        fields=fields,
+        **kwargs
+    )
+
+
+def delete_dynamic_action_definition(identifier, namespace=''):
+    return IMPL.delete_dynamic_action_definition(identifier, namespace)
+
+
+def delete_dynamic_action_definitions(**kwargs):
+    return IMPL.delete_dynamic_action_definitions(**kwargs)
+
+
+# Code sources.
+
+def get_code_source(identifier, namespace='', fields=()):
+    return IMPL.get_code_source(identifier, fields, namespace=namespace)
+
+
+def load_code_source(identifier, namespace='', fields=()):
+    return IMPL.load_code_source(identifier, fields, namespace=namespace)
+
+
+def create_code_source(values):
+    return IMPL.create_code_source(values)
+
+
+def update_code_source(identifier, values, namespace=''):
+    return IMPL.update_code_source(identifier, values, namespace=namespace)
+
+
+def get_code_sources(limit=None, marker=None, sort_keys=None,
+                     sort_dirs=None, fields=None, **kwargs):
+    return IMPL.get_code_sources(
+        limit=limit,
+        marker=marker,
+        sort_keys=sort_keys,
+        sort_dirs=sort_dirs,
+        fields=fields,
+        **kwargs
+    )
+
+
+def delete_code_source(identifier, namespace=''):
+    return IMPL.delete_code_source(identifier, namespace=namespace)
+
+
+def delete_code_sources(**kwargs):
+    return IMPL.delete_code_sources(**kwargs)
+
+
 # Action definitions.
 
 def get_action_definition_by_id(id, fields=()):
