@@ -1,5 +1,6 @@
 # Copyright 2013 - Mirantis, Inc.
 # Copyright 2015 - StackStorm, Inc.
+# Copyright 2019 - NetCracker Technology Corp.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -16,12 +17,14 @@
 import copy
 import datetime
 import json
-import mock
+from unittest import mock
+
+import sqlalchemy as sa
 
 from mistral.db.v2 import api as db_api
 from mistral.db.v2.sqlalchemy import models
-from mistral.engine.rpc_backend import rpc
 from mistral import exceptions as exc
+from mistral.rpc import clients as rpc
 from mistral.tests.unit.api import base
 from mistral.workflow import data_flow
 from mistral.workflow import states
@@ -30,6 +33,14 @@ from mistral.workflow import states
 
 RESULT = {"some": "result"}
 PUBLISHED = {"var": "val"}
+RUNTIME_CONTEXT = {
+    'triggered_by': [
+        {
+            'task_id': '123-123-123',
+            'event': 'on-success'
+        }
+    ]
+}
 
 WF_EX = models.WorkflowExecution(
     id='abc',
@@ -59,10 +70,12 @@ TASK_EX = models.TaskExecution(
     state=states.RUNNING,
     tags=['a', 'b'],
     in_context={},
-    runtime_context={},
+    runtime_context=RUNTIME_CONTEXT,
     workflow_execution_id=WF_EX.id,
     created_at=datetime.datetime(1970, 1, 1),
     updated_at=datetime.datetime(1970, 1, 1),
+    started_at=datetime.datetime(1970, 1, 1),
+    finished_at=datetime.datetime(1970, 1, 1),
     published=PUBLISHED,
     processed=True
 )
@@ -82,10 +95,12 @@ WITH_ITEMS_TASK_EX = models.TaskExecution(
     state=states.RUNNING,
     tags=['a', 'b'],
     in_context={},
-    runtime_context={},
+    runtime_context=RUNTIME_CONTEXT,
     workflow_execution_id=WF_EX.id,
     created_at=datetime.datetime(1970, 1, 1),
     updated_at=datetime.datetime(1970, 1, 1),
+    started_at=datetime.datetime(1970, 1, 1),
+    finished_at=datetime.datetime(1970, 1, 1),
     published=PUBLISHED,
     processed=True
 )
@@ -95,12 +110,16 @@ TASK = {
     'name': 'task',
     'workflow_name': 'flow',
     'workflow_id': '123e4567-e89b-12d3-a456-426655441111',
+    'tags': ['a', 'b'],
     'state': 'RUNNING',
     'workflow_execution_id': WF_EX.id,
     'created_at': '1970-01-01 00:00:00',
     'updated_at': '1970-01-01 00:00:00',
+    'started_at': '1970-01-01 00:00:00',
+    'finished_at': '1970-01-01 00:00:00',
     'result': json.dumps(RESULT),
     'published': json.dumps(PUBLISHED),
+    'runtime_context': json.dumps(RUNTIME_CONTEXT),
     'processed': True
 }
 
@@ -127,6 +146,7 @@ RERUN_TASK = {
 }
 
 MOCK_WF_EX = mock.MagicMock(return_value=WF_EX)
+TASK_EX.workflow_execution = WF_EX
 MOCK_TASK = mock.MagicMock(return_value=TASK_EX)
 MOCK_TASKS = mock.MagicMock(return_value=[TASK_EX])
 MOCK_EMPTY = mock.MagicMock(return_value=[])
@@ -134,6 +154,10 @@ MOCK_NOT_FOUND = mock.MagicMock(side_effect=exc.DBEntityNotFoundError())
 MOCK_ERROR_TASK = mock.MagicMock(return_value=ERROR_TASK_EX)
 MOCK_ERROR_ITEMS_TASK = mock.MagicMock(return_value=ERROR_ITEMS_TASK_EX)
 
+TASK_EX_WITH_PROJECT_ID = TASK_EX.get_clone()
+TASK_EX_WITH_PROJECT_ID.project_id = '<default-project>'
+TASK_EX_WITH_PROJECT_ID.workflow_execution = WF_EX
+
 
 @mock.patch.object(
     data_flow,
@@ -147,6 +171,19 @@ class TestTasksController(base.APITest):
         self.assertEqual(200, resp.status_int)
         self.assertDictEqual(TASK, resp.json)
 
+    @mock.patch.object(db_api, 'get_task_execution')
+    def test_get_operational_error(self, mocked_get):
+        mocked_get.side_effect = [
+            # Emulating DB OperationalError
+            sa.exc.OperationalError('Mock', 'mock', 'mock'),
+            TASK_EX  # Successful run
+        ]
+
+        resp = self.app.get('/v2/tasks/123')
+
+        self.assertEqual(200, resp.status_int)
+        self.assertDictEqual(TASK, resp.json)
+
     @mock.patch.object(db_api, 'get_task_execution', MOCK_NOT_FOUND)
     def test_get_not_found(self):
         resp = self.app.get('/v2/tasks/123', expect_errors=True)
@@ -162,6 +199,29 @@ class TestTasksController(base.APITest):
         self.assertEqual(1, len(resp.json['tasks']))
         self.assertDictEqual(TASK_WITHOUT_RESULT, resp.json['tasks'][0])
 
+    @mock.patch.object(db_api, 'get_task_executions')
+    def test_get_all_operational_error(self, mocked_get_all):
+        mocked_get_all.side_effect = [
+            # Emulating DB OperationalError
+            sa.exc.OperationalError('Mock', 'mock', 'mock'),
+            [TASK_EX]  # Successful run
+        ]
+
+        resp = self.app.get('/v2/tasks')
+
+        self.assertEqual(200, resp.status_int)
+
+        self.assertEqual(1, len(resp.json['tasks']))
+        self.assertDictEqual(TASK_WITHOUT_RESULT, resp.json['tasks'][0])
+
+    @mock.patch.object(db_api, 'get_task_execution',
+                       return_value=TASK_EX_WITH_PROJECT_ID)
+    def test_get_within_project_id(self, mock_get):
+        resp = self.app.get('/v2/tasks/123')
+
+        self.assertEqual(200, resp.status_int)
+        self.assertTrue('project_id' in resp.json)
+
     @mock.patch.object(db_api, 'get_task_executions', MOCK_EMPTY)
     def test_get_all_empty(self):
         resp = self.app.get('/v2/tasks')
@@ -266,6 +326,18 @@ class TestTasksController(base.APITest):
         self.assertIn('faultstring', resp.json)
         self.assertIn('execution must be in ERROR', resp.json['faultstring'])
 
+    @mock.patch.object(rpc.EngineClient, 'rerun_workflow', MOCK_WF_EX)
+    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
+    @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
+    def test_put_current_task_in_error(self):
+        params = copy.deepcopy(RERUN_TASK)
+        params['reset'] = True
+        params['env'] = '{"k1": "def"}'
+
+        resp = self.app.put_json('/v2/tasks/123', params=params)
+
+        self.assertEqual(200, resp.status_int)
+
     @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
     @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
     def test_put_invalid_state(self):
@@ -299,6 +371,20 @@ class TestTasksController(base.APITest):
         self.assertIn('faultstring', resp.json)
         self.assertIn('Only with-items task', resp.json['faultstring'])
 
+        @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
+        @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
+        def test_put_valid_state(self):
+            params = copy.deepcopy(RERUN_TASK)
+            params['state'] = states.RUNNING
+            params['reset'] = True
+
+            resp = self.app.put_json(
+                '/v2/tasks/123',
+                params=params
+            )
+
+            self.assertEqual(200, resp.status_int)
+
     @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
     @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
     def test_put_mismatch_task_name(self):
@@ -316,6 +402,22 @@ class TestTasksController(base.APITest):
         self.assertIn('faultstring', resp.json)
         self.assertIn('Task name does not match', resp.json['faultstring'])
 
+    @mock.patch.object(rpc.EngineClient, 'rerun_workflow', MOCK_WF_EX)
+    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
+    @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
+    def test_put_match_task_name(self):
+        params = copy.deepcopy(RERUN_TASK)
+        params['name'] = 'task'
+        params['reset'] = True
+
+        resp = self.app.put_json(
+            '/v2/tasks/123',
+            params=params,
+            expect_errors=True
+        )
+
+        self.assertEqual(200, resp.status_int)
+
     @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
     @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
     def test_put_mismatch_workflow_name(self):
