@@ -19,8 +19,9 @@ from mistral.db.v2 import api as db_api
 from mistral.lang.v2 import tasks as tasks_lang
 from mistral.services import workflows as wf_service
 from mistral.tests.unit.engine import base
-from mistral import utils
 from mistral.workflow import states
+from mistral_lib import actions as actions_base
+from mistral_lib import utils
 
 
 # Use the set_default method to set value otherwise in certain test cases
@@ -28,6 +29,22 @@ from mistral.workflow import states
 cfg.CONF.set_default('auth_enable', False, group='pecan')
 
 
+class ActionWithExceptionInInit(actions_base.Action):
+    def __init__(self, aaa):
+        super(ActionWithExceptionInInit, self).__init__()
+
+        if aaa != "bbb":
+            raise Exception("Aaa doesn't equal to bbb")
+
+        self.aaa = aaa
+
+    def run(self, context):
+        return actions_base.Result(data=self.aaa)
+
+    def test(self):
+        raise NotImplementedError
+
+
 class JoinEngineTest(base.EngineTestCase):
     def test_full_join_simple(self):
         wf_text = """---
@@ -235,7 +252,6 @@ class JoinEngineTest(base.EngineTestCase):
         # uncertainty of its running in parallel with task3.
         self.await_task_success(task4.id)
 
-        self.assertEqual(states.RUNNING, wf_ex.state)
         self.assertEqual(states.SUCCESS, task1.state)
         self.assertEqual(states.SUCCESS, task2.state)
 
@@ -873,18 +889,11 @@ class JoinEngineTest(base.EngineTestCase):
             state=states.WAITING
         )
 
-        calls = db_api.get_delayed_calls()
-
-        mtd_name = 'mistral.engine.task_handler._refresh_task_state'
-
-        cnt = sum([1 for c in calls if c.target_method_name == mtd_name])
-
-        # There can be 2 calls with different value of 'processing' flag.
-        self.assertTrue(cnt == 1 or cnt == 2)
-
         # Stop the workflow.
         self.engine.stop_workflow(wf_ex.id, state=states.CANCELLED)
 
+        mtd_name = 'mistral.engine.task_handler._refresh_task_state'
+
         self._await(
             lambda:
             len(db_api.get_delayed_calls(target_method_name=mtd_name)) == 0
@@ -931,18 +940,11 @@ class JoinEngineTest(base.EngineTestCase):
             state=states.WAITING
         )
 
-        calls = db_api.get_delayed_calls()
-
-        mtd_name = 'mistral.engine.task_handler._refresh_task_state'
-
-        cnt = sum([1 for c in calls if c.target_method_name == mtd_name])
-
-        # There can be 2 calls with different value of 'processing' flag.
-        self.assertTrue(cnt == 1 or cnt == 2)
-
         # Stop the workflow.
         db_api.delete_workflow_execution(wf_ex.id)
 
+        mtd_name = 'mistral.engine.task_handler._refresh_task_state'
+
         self._await(
             lambda:
             len(db_api.get_delayed_calls(target_method_name=mtd_name)) == 0
@@ -1308,3 +1310,97 @@ class JoinEngineTest(base.EngineTestCase):
         self.assertEqual(states.SUCCESS, task1.state)
         self.assertEqual(states.SUCCESS, task2.state)
         self.assertEqual(states.SUCCESS, task3.state)
+
+    def test_join_last_inbound_indirect_error(self):
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          tasks:
+            task1:
+              action: std.noop
+              on-success:
+                - join_task
+
+            task2:
+              action: std.fail
+              wait-before: 2
+              on-success:
+                - task3
+
+            task3:
+              action: std.noop
+              on-success:
+                - join_task
+
+            join_task:
+              join: all
+        """
+
+        wf_service.create_workflows(wf_text)
+
+        wf_ex = self.engine.start_workflow('wf')
+
+        self.await_workflow_error(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
+
+            self.assertEqual(3, len(task_execs))
+
+            self._assert_single_item(
+                task_execs,
+                name='task1',
+                state=states.SUCCESS
+            )
+            self._assert_single_item(
+                task_execs,
+                name='task2',
+                state=states.ERROR
+            )
+            self._assert_single_item(
+                task_execs,
+                name='join_task',
+                state=states.ERROR
+            )
+
+    def test_join_task_with_input_error(self):
+        self.register_action_class(
+            'my_action',
+            ActionWithExceptionInInit
+        )
+
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          type: direct
+
+          tasks:
+            join_task:
+              action: my_action aaa="aaa"
+              join: all
+
+            task1:
+              on-success: join_task
+
+            task2:
+              on-success: join_task
+        """
+
+        wf_service.create_workflows(wf_text)
+
+        wf_ex = self.engine.start_workflow('wf')
+
+        self.await_workflow_error(wf_ex.id, timeout=5)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            t_execs = wf_ex.task_executions
+
+        self._assert_single_item(t_execs, name='task1', state=states.SUCCESS)
+        self._assert_single_item(t_execs, name='task2', state=states.SUCCESS)
+        self._assert_single_item(t_execs, name='join_task', state=states.ERROR)
