@@ -1,4 +1,5 @@
 # Copyright 2015 - Mirantis, Inc.
+# Copyright 2020 - NetCracker Technology Corp.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -15,18 +16,20 @@
 from oslo_log import log as logging
 from osprofiler import profiler
 
+from mistral.db.v2 import api as db_api
 from mistral import exceptions as exc
 from mistral import expressions as expr
-from mistral import utils
 from mistral.workflow import base
 from mistral.workflow import commands
 from mistral.workflow import data_flow
-from mistral.workflow import lookup_utils
 from mistral.workflow import states
+from mistral_lib import utils
 
 
 LOG = logging.getLogger(__name__)
 
+MAX_SEARCH_DEPTH = 5
+
 
 class DirectWorkflowController(base.WorkflowController):
     """'Direct workflow' controller.
@@ -44,37 +47,37 @@ class DirectWorkflowController(base.WorkflowController):
     __workflow_type__ = "direct"
 
     def _get_upstream_task_executions(self, task_spec):
-        return list(
-            filter(
-                lambda t_e: self._is_upstream_task_execution(task_spec, t_e),
-                lookup_utils.find_task_executions_by_specs(
-                    self.wf_ex.id,
-                    self.wf_spec.find_inbound_task_specs(task_spec)
-                )
+        t_specs_names = [t_spec.get_name() for t_spec in
+                         self.wf_spec.find_inbound_task_specs(task_spec)]
+
+        if not t_specs_names:
+            return []
+
+        if not task_spec.get_join():
+            return self._get_task_executions(
+                name=t_specs_names[0],  # not a join, has just one parent
+                state={'in': (states.SUCCESS, states.ERROR, states.CANCELLED)},
+                processed=True
             )
+
+        t_execs_candidates = self._get_task_executions(
+            name={'in': t_specs_names},
+            state={'in': (states.SUCCESS, states.ERROR, states.CANCELLED)},
         )
 
-    def _is_upstream_task_execution(self, t_spec, t_ex_candidate):
-        if not states.is_completed(t_ex_candidate.state):
-            return False
+        t_execs = []
+        for t_ex in t_execs_candidates:
+            if task_spec.get_name() in [t[0] for t in t_ex.next_tasks]:
+                t_execs.append(t_ex)
 
-        if not t_spec.get_join():
-            return t_ex_candidate.processed
-
-        induced_state, _, _ = self._get_induced_join_state(
-            self.wf_spec.get_tasks()[t_ex_candidate.name],
-            self._find_task_execution_by_name(t_ex_candidate.name),
-            t_spec
-        )
-
-        return induced_state == states.RUNNING
+        return t_execs
 
     def _find_next_commands(self, task_ex=None):
         cmds = super(DirectWorkflowController, self)._find_next_commands(
             task_ex
         )
 
-        # Checking if task_ex is empty here is a serious optimization here
+        # Checking if task_ex is empty is a serious optimization here
         # because 'self.wf_ex.task_executions' leads to initialization of
         # the entire collection which in case of highly-parallel workflows
         # may be very expensive.
@@ -105,6 +108,10 @@ class DirectWorkflowController(base.WorkflowController):
             for t_s in self.wf_spec.find_start_tasks()
         ]
 
+    @profiler.trace(
+        'direct-wf-controller-find-next-commands-for-task',
+        hide_args=True
+    )
     def _find_next_commands_for_task(self, task_ex):
         """Finds next commands based on the state of the given task.
 
@@ -117,13 +124,13 @@ class DirectWorkflowController(base.WorkflowController):
 
         ctx = data_flow.evaluate_task_outbound_context(task_ex)
 
-        for t_n, params, event_name in self._find_next_tasks(task_ex, ctx=ctx):
+        for t_n, params, event_name in self._find_next_tasks(task_ex, ctx):
             t_s = self.wf_spec.get_tasks()[t_n]
 
             if not (t_s or t_n in commands.ENGINE_CMD_CLS):
                 raise exc.WorkflowException("Task '%s' not found." % t_n)
             elif not t_s:
-                t_s = self.wf_spec.get_tasks()[task_ex.name]
+                t_s = self.wf_spec.get_task(task_ex.name)
 
             triggered_by = [
                 {
@@ -139,7 +146,8 @@ class DirectWorkflowController(base.WorkflowController):
                 t_s,
                 ctx,
                 params=params,
-                triggered_by=triggered_by
+                triggered_by=triggered_by,
+                handles_error=(event_name == 'on-error')
             )
 
             self._configure_if_join(cmd)
@@ -151,7 +159,7 @@ class DirectWorkflowController(base.WorkflowController):
         return cmds
 
     def _configure_if_join(self, cmd):
-        if not isinstance(cmd, commands.RunTask):
+        if not isinstance(cmd, (commands.RunTask, commands.RunExistingTask)):
             return
 
         if not cmd.task_spec.get_join():
@@ -163,15 +171,23 @@ class DirectWorkflowController(base.WorkflowController):
     def _get_join_unique_key(self, cmd):
         return 'join-task-%s-%s' % (self.wf_ex.id, cmd.task_spec.get_name())
 
+    def rerun_tasks(self, task_execs, reset=True):
+        cmds = super(DirectWorkflowController, self).rerun_tasks(
+            task_execs,
+            reset
+        )
+
+        for cmd in cmds:
+            self._configure_if_join(cmd)
+
+        return cmds
+
     # TODO(rakhmerov): Need to refactor this method to be able to pass tasks
     # whose contexts need to be merged.
     def evaluate_workflow_final_context(self):
         ctx = {}
 
         for batch in self._find_end_task_executions_as_batches():
-            if not batch:
-                break
-
             for t_ex in batch:
                 ctx = utils.merge_dicts(
                     ctx,
@@ -191,68 +207,82 @@ class DirectWorkflowController(base.WorkflowController):
 
         return self._get_join_logical_state(task_spec)
 
+    def find_indirectly_affected_task_executions(self, t_name):
+        all_joins = {task_spec.get_name()
+                     for task_spec in self.wf_spec.get_tasks()
+                     if task_spec.get_join()}
+
+        t_execs_cache = {
+            t_ex.name: t_ex for t_ex in self._get_task_executions(
+                fields=('id', 'name'),
+                name={'in': all_joins}
+            )
+        } if all_joins else {}
+
+        visited_task_names = set()
+        clauses = self.wf_spec.find_outbound_task_names(t_name)
+
+        res = set()
+
+        while clauses:
+            visited_task_names.add(t_name)
+            t_name = clauses.pop()
+
+            # Handle cycles.
+            if t_name in visited_task_names:
+                continue
+
+            # Encountered an engine command.
+            if not self.wf_spec.get_tasks()[t_name]:
+                continue
+
+            if t_name in all_joins and t_name in t_execs_cache:
+                res.add(t_execs_cache[t_name])
+                continue
+
+            clauses.update(self.wf_spec.find_outbound_task_names(t_name))
+
+        return res
+
     def is_error_handled_for(self, task_ex):
+        # TODO(rakhmerov): The method works in a different way than
+        # all_errors_handled(). It doesn't evaluate expressions under
+        # "on-error" clause.
         return bool(self.wf_spec.get_on_error_clause(task_ex.name))
 
     def all_errors_handled(self):
-        for t_ex in lookup_utils.find_error_task_executions(self.wf_ex.id):
-            ctx_view = data_flow.ContextView(
-                data_flow.evaluate_task_outbound_context(t_ex),
-                data_flow.get_workflow_environment_dict(self.wf_ex),
-                self.wf_ex.context,
-                self.wf_ex.input
-            )
+        cnt = db_api.get_task_executions_count(
+            workflow_execution_id=self.wf_ex.id,
+            state=states.ERROR,
+            error_handled=False
+        )
 
-            tasks_on_error = self._find_next_tasks_for_clause(
-                self.wf_spec.get_on_error_clause(t_ex.name),
-                ctx_view
-            )
-
-            if not tasks_on_error:
-                return False
-
-        return True
+        return cnt == 0
 
     def _find_end_task_executions_as_batches(self):
-        def is_end_task(t_ex):
-            try:
-                return not self._has_outbound_tasks(t_ex)
-            except exc.MistralException:
-                # If some error happened during the evaluation of outbound
-                # tasks we consider that the given task is an end task.
-                # Due to this output-on-error could reach the outbound context
-                # of given task also.
-                return True
-
-        batches = lookup_utils.find_completed_task_executions_as_batches(
-            self.wf_ex.id
+        batches = db_api.get_completed_task_executions_as_batches(
+            workflow_execution_id=self.wf_ex.id,
+            has_next_tasks=False
         )
 
         for batch in batches:
-            yield list(
-                filter(is_end_task, batch)
-            )
+            yield batch
 
-    def _has_outbound_tasks(self, task_ex):
-        # In order to determine if there are outbound tasks we just need
-        # to calculate next task names (based on task outbound context)
-        # and remove all engine commands. To do the latter it's enough to
-        # check if there's a corresponding task specification for a task name.
-        return bool([
-            t_name for t_name in self._find_next_task_names(task_ex)
-            if self.wf_spec.get_tasks()[t_name]
-        ])
+    def may_complete_workflow(self, task_ex):
+        res = super(DirectWorkflowController, self).may_complete_workflow(
+            task_ex
+        )
 
-    def _find_next_task_names(self, task_ex):
-        return [t[0] for t in self._find_next_tasks(task_ex)]
+        return res and not task_ex.has_next_tasks
 
-    def _find_next_tasks(self, task_ex, ctx=None):
-        t_state = task_ex.state
-        t_name = task_ex.name
+    @profiler.trace('direct-wf-controller-find-next-tasks', hide_args=True)
+    def _find_next_tasks(self, task_ex, ctx):
+        t_n = task_ex.name
+        t_s = task_ex.state
 
         ctx_view = data_flow.ContextView(
             data_flow.get_current_task_dict(task_ex),
-            ctx or data_flow.evaluate_task_outbound_context(task_ex),
+            ctx,
             data_flow.get_workflow_environment_dict(self.wf_ex),
             self.wf_ex.context,
             self.wf_ex.input
@@ -261,54 +291,33 @@ class DirectWorkflowController(base.WorkflowController):
         # [(task_name, params, 'on-success'|'on-error'|'on-complete'), ...]
         result = []
 
-        def process_clause(clause, event_name):
-            task_tuples = self._find_next_tasks_for_clause(clause, ctx_view)
+        if t_s == states.ERROR:
+            for name, cond, params in self.wf_spec.get_on_error_clause(t_n):
+                if not cond or expr.evaluate(cond, ctx_view):
+                    params = expr.evaluate_recursively(params, ctx_view)
 
-            for t in task_tuples:
-                result.append((t[0], t[1], event_name))
+                    result.append((name, params, 'on-error'))
 
-        if t_state == states.SUCCESS:
-            process_clause(
-                self.wf_spec.get_on_success_clause(t_name),
-                'on-success'
-            )
-        elif t_state == states.ERROR:
-            process_clause(
-                self.wf_spec.get_on_error_clause(t_name),
-                'on-error'
-            )
+        if t_s == states.SUCCESS:
+            for name, cond, params in self.wf_spec.get_on_success_clause(t_n):
+                if not cond or expr.evaluate(cond, ctx_view):
+                    params = expr.evaluate_recursively(params, ctx_view)
 
-        if states.is_completed(t_state) and not states.is_cancelled(t_state):
-            process_clause(
-                self.wf_spec.get_on_complete_clause(t_name),
-                'on-complete'
-            )
+                    result.append((name, params, 'on-success'))
+
+        if states.is_completed(t_s) and not states.is_cancelled(t_s):
+            for name, cond, params in self.wf_spec.get_on_complete_clause(t_n):
+                if not cond or expr.evaluate(cond, ctx_view):
+                    params = expr.evaluate_recursively(params, ctx_view)
+
+                    result.append((name, params, 'on-complete'))
 
         return result
 
-    @staticmethod
-    def _find_next_tasks_for_clause(clause, ctx):
-        """Finds next tasks names.
-
-         This method finds next task(command) base on given {name: condition}
-         dictionary.
-
-        :param clause: Tuple (task_name, condition, parameters) taken from
-            'on-complete', 'on-success' or 'on-error' clause.
-        :param ctx: Context that clause expressions should be evaluated
-            against of.
-        :return: List of task(command) names.
-        """
-        if not clause:
-            return []
-
-        return [
-            (t_name, expr.evaluate_recursively(params, ctx))
-            for t_name, condition, params in clause
-            if not condition or expr.evaluate(condition, ctx)
-        ]
-
-    @profiler.trace('direct-wf-controller-get-join-logical-state')
+    @profiler.trace(
+        'direct-wf-controller-get-join-logical-state',
+        hide_args=True
+    )
     def _get_join_logical_state(self, task_spec):
         """Evaluates logical state of 'join' task.
 
@@ -323,9 +332,6 @@ class DirectWorkflowController(base.WorkflowController):
         # TODO(rakhmerov): We need to use task_ex instead of task_spec
         # in order to cover a use case when there's more than one instance
         # of the same 'join' task in a workflow.
-        # TODO(rakhmerov): In some cases this method will be expensive because
-        # it uses a multi-step recursive search. We need to optimize it moving
-        # forward (e.g. with Workflow Execution Graph).
 
         join_expr = task_spec.get_join()
 
@@ -334,13 +340,20 @@ class DirectWorkflowController(base.WorkflowController):
         if not in_task_specs:
             return base.TaskLogicalState(states.RUNNING)
 
+        t_execs_cache = self._prepare_task_executions_cache(task_spec)
+
         # List of tuples (task_name, task_ex, state, depth, event_name).
         induced_states = []
 
         for t_s in in_task_specs:
-            t_ex = self._find_task_execution_by_name(t_s.get_name())
+            t_ex = t_execs_cache[t_s.get_name()]
 
-            tup = self._get_induced_join_state(t_s, t_ex, task_spec)
+            tup = self._get_induced_join_state(
+                t_s,
+                t_ex,
+                task_spec,
+                t_execs_cache
+            )
 
             induced_states.append(
                 (
@@ -439,12 +452,19 @@ class DirectWorkflowController(base.WorkflowController):
     # TODO(rakhmerov): Method signature is incorrect given that
     # we may have multiple task executions for a task. It should
     # accept inbound task execution rather than a spec.
+    @profiler.trace(
+        'direct-wf-controller-get-induced-join-state',
+        hide_args=True
+    )
     def _get_induced_join_state(self, in_task_spec, in_task_ex,
-                                join_task_spec):
+                                join_task_spec, t_execs_cache):
         join_task_name = join_task_spec.get_name()
 
         if not in_task_ex:
-            possible, depth = self._possible_route(in_task_spec)
+            possible, depth = self._possible_route(
+                in_task_spec,
+                t_execs_cache
+            )
 
             if possible:
                 return states.WAITING, depth, None
@@ -454,47 +474,80 @@ class DirectWorkflowController(base.WorkflowController):
         if not states.is_completed(in_task_ex.state):
             return states.WAITING, 1, None
 
-        # [(task name, params, event name), ...]
-        next_tasks_tuples = self._find_next_tasks(in_task_ex)
+        # [(task name, event name), ...]
+        next_tasks_tuples = in_task_ex.next_tasks or []
 
-        next_tasks_dict = {tup[0]: tup[2] for tup in next_tasks_tuples}
+        next_tasks_dict = {tup[0]: tup[1] for tup in next_tasks_tuples}
 
         if join_task_name not in next_tasks_dict:
             return states.ERROR, 1, "not triggered"
 
         return states.RUNNING, 1, next_tasks_dict[join_task_name]
 
-    def _find_task_execution_by_name(self, t_name):
-        # Note: in case of 'join' completion check it's better to initialize
-        # the entire task_executions collection to avoid too many DB queries.
-
-        t_execs = lookup_utils.find_task_executions_by_name(
-            self.wf_ex.id,
-            t_name
-        )
-
-        # TODO(rakhmerov): Temporary hack. See the previous comment.
-        return t_execs[-1] if t_execs else None
-
-    def _possible_route(self, task_spec, depth=1):
+    def _possible_route(self, task_spec, t_execs_cache, depth=1):
         in_task_specs = self.wf_spec.find_inbound_task_specs(task_spec)
 
         if not in_task_specs:
             return True, depth
 
         for t_s in in_task_specs:
-            t_ex = self._find_task_execution_by_name(t_s.get_name())
+            if t_s.get_name() not in t_execs_cache:
+                t_execs_cache.update(
+                    self._prepare_task_executions_cache(task_spec)
+                )
+
+            t_ex = t_execs_cache.get(t_s.get_name())
 
             if not t_ex:
-                possible, depth = self._possible_route(t_s, depth + 1)
+                possible, depth = self._possible_route(
+                    t_s,
+                    t_execs_cache,
+                    depth + 1
+                )
 
                 if possible:
                     return True, depth
             else:
                 t_name = task_spec.get_name()
 
-                if (not states.is_completed(t_ex.state) or
-                        t_name in self._find_next_task_names(t_ex)):
+                if not states.is_completed(t_ex.state):
+                    return True, depth
+
+                if t_name in [t[0] for t in t_ex.next_tasks]:
                     return True, depth
 
         return False, depth
+
+    def _find_all_parent_task_names(self, task_spec, depth=1):
+        if depth == MAX_SEARCH_DEPTH:
+            return {task_spec.get_name()}
+
+        in_task_specs = self.wf_spec.find_inbound_task_specs(task_spec)
+
+        if not in_task_specs:
+            return {task_spec.get_name()}
+
+        names = set()
+        for t_s in in_task_specs:
+            names.update(self._find_all_parent_task_names(t_s, depth + 1))
+
+        if depth > 1:
+            names.add(task_spec.get_name())
+
+        return names
+
+    def _prepare_task_executions_cache(self, task_spec):
+        names = self._find_all_parent_task_names(task_spec)
+
+        t_execs_cache = {
+            t_ex.name: t_ex for t_ex in self._get_task_executions(
+                fields=('id', 'name', 'state', 'next_tasks'),
+                name={'in': names}
+            )
+        } if names else {}  # don't perform a db request if 'names' are empty
+
+        for name in names:
+            if name not in t_execs_cache:
+                t_execs_cache[name] = None
+
+        return t_execs_cache
